/* $Id: altq_cbq.c,v 1.1 1998/06/08 09:56:47 kjc Exp kjc $ */
/*
 * Copyright (C) 1997, 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "opt_altq.h"

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <math.h>

#include "command.h"
#include "mbuf.h"
#include "filter.h"
#include "timer.h"
#include "defs.h"
#include "log.h"
#include "loadalias.h"
#include "vars.h"

#include "altq_ppp.h"
#include "altq_classq.h"
#include "altq_rmclass.h"
#include "altq_rmclass_debug.h"
#include "altq_cbq.h"
#ifdef CBQ_RED
#include "altq_red.h"
#endif
#ifdef CBQ_PROBE
#include "../cbqmonitor/cbqmon.h"
#endif

static int cbq_enqueue(struct mbuf *);
static struct mbuf *cbq_dequeue(void);
static int cbq_account(struct mbuf *);
static int cbq_set_params(int, int, int, int, u_int *, u_int *, int *,
			  u_int *, int *, int *);

struct rm_ifdat *ifd = NULL;
struct rm_class *cbq_classes[CBQ_MAXCLASS];

int cbq_enable(void)
{
	int rval = 0;
	
	if (ifd == NULL) {
		LogPrintf(LogWARN, "cbq: no interface!\n");
		return (-1);
	}
	if (ifd->root_ == NULL) {
		LogPrintf(LogWARN, "cbq: no root class!\n");
		return (-1);
	}
	if (ifd->default_ == NULL) {
		LogPrintf(LogWARN, "cbq: no default class!\n");
		return (-1);
	}

	/* if not attached yet, attach it */
	if (altqp->altq_type == ALTQT_NONE)
		rval = altq_attach(cbq_enqueue, cbq_dequeue, cbq_account,
				   ALTQT_CBQ);
	else if (altqp->altq_type != ALTQT_CBQ) {
		LogPrintf(LogERROR, "cbq: can't attach!\n");
		return (-1);
	}

	if (rval == 0)
		rval = altq_enable(ALTQT_CBQ);

	if (rval != 0)
		LogPrintf(LogERROR, "can't enable CBQ!\n");
	else if (VarTerm)
		fprintf(VarTerm, "cbq enabled\n");
	
	return (rval);
}

int cbq_disable(void)
{
	struct rm_class	*cl;
	int 		i, rval;
    
	for (i = 0; i < CBQ_MAXCLASS; i++)
		if ((cl = cbq_classes[i]) != NULL)
			rmc_dropall(cl);

	rval = altq_disable(ALTQT_CBQ);

	if (rval != 0)
		LogPrintf(LogERROR, "can't disable CBQ!\n");
	else if (VarTerm)
		fprintf(VarTerm, "cbq disabled\n");

#if 1
	/*
	 * at this moment, CBQ is the only discipline supported.
	 * so, don't detach it and go into the accounting mode.
	 */
	altqp->altq_flags |= ALTQF_ACCOUNTING;
#else
	if (rval == 0)
		rval = altq_detach(ALTQT_CBQ);
#endif

	return (rval);
}

int
cbq_ifinit(bandwidth, workconserving)
	int	bandwidth, workconserving;
{
	double ns_per_byte;

        ns_per_byte = 1.0 / (double)bandwidth * 8 * RM_NS_PER_SEC;

	ifd = rmc_init((u_int)ns_per_byte, RM_MAXQUEUED, workconserving);

	ifd->bandwidth_ = bandwidth;

	return (0);
}

struct rm_class *
cbq_class_create(cname, parent_name, borrow, pri, pbandwidth,
		 maxburst, minburst, maxq, pktsize, flags)
	char	*cname, *parent_name;
	int	borrow, pbandwidth, maxburst, minburst, maxq, pktsize, flags;
{
	struct rm_class *parent_cl, *borrow_cl, *cl;
	int	minidle;
	u_int	nsec_per_byte, maxidle, offtime;
	int	i;

	if (parent_name != NULL)
		parent_cl = cbq_class_find(parent_name);
	else
		parent_cl = NULL;
	if (parent_cl == NULL) {
		if (ifd->root_ != NULL) {
			LogPrintf(LogERROR,
				  "no parent class specified for %s!", cname);
			return (NULL);
		}
	}

	if (borrow)
		borrow_cl = parent_cl;
	else
		borrow_cl = NULL;

	cbq_set_params(pbandwidth, ifd->bandwidth_, maxburst, minburst,
		       &nsec_per_byte, &maxidle, &minidle, &offtime,
		       &maxq, &pktsize);

#if 1
	if (pbandwidth == 0 && borrow == NULL)
		/* filter out this class by setting queue size to zero */
		maxq = 0;
#endif

	cl = rmc_newclass(pri, ifd, nsec_per_byte,
			  rmc_delay_action, maxq, parent_cl, borrow_cl,
			  maxidle, minidle, offtime,
			  pktsize, flags);
	if (cl == NULL) {
		LogPrintf(LogERROR, "can't create class %s!", cname);
		return (NULL);
	}
	strcpy(cl->name_, cname);

	if (parent_cl == NULL) {
		ifd->root_ = cl;
		ifd->root_->depth_ = 0;
	}
	else if (flags & CBQCLF_DEFCLASS)
		ifd->default_ = cl;

	for (i = 0; i < CBQ_MAXCLASS; i++)
		if (cbq_classes[i] == NULL) {
			cbq_classes[i] = cl;
			break;
		}

	return (cl);
}

void
cbq_class_delete(cl)
	struct rm_class *cl;
{
	int i;
	
	/* if the class has filters, delete them first. */
	DeleteCBQfilter(cl);

	rmc_delete_class(ifd, cl);

	for (i = 0; i < CBQ_MAXCLASS; i++)
		if (cbq_classes[i] == cl) {
			cbq_classes[i] = NULL;
			break;
		}
}

struct rm_class *
cbq_class_find(char *cname)
{
    struct rm_class *cl, *clh;
    int i;

    for (i = 0; i < RM_MAXPRIO; i++)
	if ((clh = ifd->active_[i]) != NULL) {
	    cl = clh;
	    do {
		if (strcmp(cname, cl->name_) == 0)
		    return (cl);
		cl = cl->peer_;
	    } while (cl != clh);
	}
    return (NULL);
}

static int
cbq_enqueue(m)
	struct mbuf *m;
{
	struct rm_class *cl;
	struct filterent *fp;

	fp = m->fp;
	if (fp == NULL || (fp->action & A_CBQ) == 0)
		cl = ifd->default_;
	else
		cl = fp->owner;
	
	if (rmc_queue_packet(cl, m) == 0) {
		/* successfully queued. */
		++ifd->qlen_;
		return (0);
	}
    			
	/* drop occurred.  some mbuf was freed in rmc_queue_packet. */
	return (ENOBUFS);
}

static struct mbuf *
cbq_dequeue(void)
{
	struct mbuf 	*m;

	if (ifd->qlen_ == 0)
		return (NULL);
	if ((m = rmc_dequeue_next(ifd)) != NULL) {
		--ifd->qlen_;  /* decrement # of packets in cbq */
	
		/* Update the class. */
		rmc_update_class_util(ifd);
	}
	return (m);
}

static int
cbq_account(m)
	struct mbuf *m;
{
	struct rm_class *cl;
	struct filterent *fp;
	int len;

	fp = m->fp;
	if (fp == NULL || (fp->action & A_CBQ) == 0)
		cl = ifd->default_;
	else
		cl = fp->owner;

	len = m_pktlen(m);
	
	cl->stats_.npackets++;
	cl->stats_.nbytes += len;

	ifd->root_->stats_.npackets++;
	ifd->root_->stats_.nbytes += len;

	return (0);
}


static int
cbq_set_params(pbandwidth, link_bandwidth, maxburst, minburst,
	       p_ns_per_byte, p_maxidle, p_minidle, p_offtime,
	       p_maxq, p_pktsize)
	int	pbandwidth, link_bandwidth, maxburst, minburst;
	u_int	*p_ns_per_byte, *p_maxidle, *p_offtime;
	int	*p_minidle, *p_maxq, *p_pktsize;
{
	double	ns_per_byte;
	double	maxidle, min_maxidle, offtime;
	double	minidle;
	double	ptime;		/* target interpacket time */
	double	f;		/* fraction of link bandwidth
				   allocated to the calss */
	double	pkttime;	/* physical packet time (ns) */
	double	g;		/* 1 - 2^RM_FILTER_GAIN */
	double	g_maxb;		/* g^maxburst */
	double	g_minb;		/* g^minburst */
	int	max_pktsize, avg_pktsize;
	int	maxq;

	/* import parameters */
	maxq = *p_maxq;
	avg_pktsize = *p_pktsize;
	
	/* if packet size isn't set, set MTU */
	max_pktsize = altqp->altq_mtu;
	if (avg_pktsize == 0)
		avg_pktsize = max_pktsize;
	f = (double)pbandwidth / 100.0;
        ns_per_byte = 1.0 / ((double)link_bandwidth * f) * 8 * RM_NS_PER_SEC;

	pkttime = (double)avg_pktsize * (double)ns_per_byte * f;
	ptime = pkttime * (1.0/f - 1.0);

	g = 1.0 - 1.0/(double)RM_POWER;
	g_maxb = pow(g, (double)maxburst);
	g_minb = pow(g, (double)minburst);

	maxidle = ptime * (1.0/f - 1.0) * (1.0 - g_maxb)/g_maxb;
	min_maxidle = ptime * (1.0 - g);
	if (maxidle < min_maxidle)
		maxidle = min_maxidle;

	offtime = ptime * (1.0/f - 1.0);
	if (minburst > 0)
		offtime *=  1.0 + 1.0/(1.0 - g) * (1.0 - g_minb)/g_minb;
#if 1
	minidle = ptime * 2 / (1 - RM_POWER);
#else
	minidle = -((double)max_pktsize * ns_per_byte);
#endif
	
#if 1
	if (ns_per_byte * (double)max_pktsize > (double)INT_MAX) {
		/* this cause integer overflow in kernel! */
		if (pbandwidth != 0)
			LogPrintf(LogWARN, "warning: class is too slow!!");
		ns_per_byte = (double)(INT_MAX / max_pktsize);
	}
#endif
	if (altq_debug && VarTerm) {
		int packet_time;
		fprintf(VarTerm, "cbq_set_params: maxburst=%d,minburst=%d,pkt_size=%d\n",
		       maxburst, minburst, avg_pktsize);
		fprintf(VarTerm, "  nsPerByte=%.2f ns, link's nsPerByte=%.2f, p=%.3f\n",
		       ns_per_byte,
		       1.0 / (double)link_bandwidth * 8 * RM_NS_PER_SEC, f);
		fprintf(VarTerm, "  maxilde=%.2f, minidle=%.2f, offtime=%.2f\n",
		       maxidle/1000, minidle/1000, offtime/1000);
		packet_time = avg_pktsize * (int)ns_per_byte / 1000;
		fprintf(VarTerm, "  packet time=%d [us]\n", packet_time);
		if (maxburst * packet_time < 20000) {
			fprintf(VarTerm, "warning: maxburst smaller than timer granularity!\n");
			fprintf(VarTerm, "         maxburst=%d, packet_time=%d [us]\n",
			       maxburst, packet_time);
		}
	}

	if (maxq == 0)
		maxq = 30;
	if (maxq < 4) {
		LogPrintf(LogWARN, "warning: maxq (%d) is too small. set to %d", (int)maxq, 4);
		maxq = 4;
	}
	if (1) {
		if ((int)maxq < maxburst)
			LogPrintf(LogWARN, "warning: maxq (%d) is smaller than maxburst(%d)\n",
			    maxq, maxburst);
		else if (maxq > 100)
			LogPrintf(LogWARN, "warning: maxq %d too large\n", maxq);
		if (altq_debug && VarTerm)
			fprintf(VarTerm, "  maxq=%d\n", maxq);
	}

	/* maxidle, minidle and offtime are scaled */
	maxidle = maxidle * 8.0 / ns_per_byte * pow(2, RM_FILTER_GAIN);
	minidle = minidle * pow(2, RM_FILTER_GAIN);
 	offtime = offtime * 8.0 / ns_per_byte * pow(2, RM_FILTER_GAIN);
	maxidle = maxidle / 1000.0;
	minidle = minidle / 1000.0;
	offtime = offtime / 1000.0;

	*p_ns_per_byte	= (u_int)ns_per_byte;
	*p_maxidle	= (u_int)maxidle;
	*p_minidle	= (int)minidle;
	*p_offtime	= (u_int)offtime;
	*p_maxq		= (int)maxq;
	*p_pktsize	= (int)avg_pktsize;

	return (0);
}


struct pppTimer cbqStatTimer;

int
cbq_status(int interval)
{
    struct rm_class *cl;
    int i;

    if (interval < 0) {
	    StopTimer(&cbqStatTimer);
	    return (0);
    }
    
    if (VarTerm) {
	fprintf(VarTerm, "\ncbq status: %s\n",
		ALTQ_IS_ON() ? "enabled" : "disabled");
	for (i = 0; i < CBQ_MAXCLASS; i++)
	    if ((cl = cbq_classes[i]) != NULL) {
		fprintf(VarTerm, "class %s: nsPerByte:%d qlen:%d/%d\n",
			cl->name_, cl->ns_per_byte_, cl->q_->qlen_, cl->q_->qlim_);
		fprintf(VarTerm, "  pri:%d depth:%d"
			" offtime:%d[us] wrr_allot:%d\n",
			cl->pri_, cl->depth_,
			cl->offtime_, cl->w_allotment_);
		fprintf(VarTerm, "  npackets:%d drops:%d borrows:%d delays:%d\n",
			cl->stats_.npackets, cl->stats_.drops, 
			cl->stats_.borrows, cl->stats_.delays);
		fprintf(VarTerm, "  avgidle:%d[us] (max:%d min:%d)\n",
			cl->avgidle_ >> RM_FILTER_GAIN,
			cl->maxidle_ >> RM_FILTER_GAIN,
			cl->minidle_ >> RM_FILTER_GAIN);
		if (cl->sleeping_) {
			fprintf(VarTerm, "sleeping: undertime:%ld:%ld "
				"last:%ld:%ld overtime:%ld:%ld (ifnow%ld:%ld)\n",
				cl->undertime_.tv_sec,cl->undertime_.tv_usec,
				cl->last_.tv_sec,cl->last_.tv_usec,
				cl->overtime_.tv_sec,cl->overtime_.tv_usec,
				ifd->ifnow_.tv_sec,ifd->ifnow_.tv_usec);
		}
#ifdef CBQ_RED
		if (cl->red_ != NULL)
			fprintf(VarTerm, "  red q_avg:%.2f"
				"(forced: %qd, early: %qd marked: %qd)\n",
				(double)cl->red_->red_avg /
				(double)(1 << cl->red_->red_avgshift),
				cl->red_->red_stats.drop_forced,
				cl->red_->red_stats.drop_unforced,
				cl->red_->red_stats.marked_packets);
#endif /* CBQ_RED */
	    }
    }

    if (interval > 0) {
	    cbqStatTimer.load = interval * SECTICKS;
	    cbqStatTimer.func = (void(*)(void *))cbq_status;
	    cbqStatTimer.arg = (void *)interval;
	    StartTimer(&cbqStatTimer);
    }

    return (0);
}

#ifdef CBQ_PROBE
/*
 * cbqprobe function for use with cbqmonitor
 */
static int sockfd = -1;
static struct sockaddr_in view_addr;

static char buffer[sizeof(struct cbqm_info)];

/* convert a string name to struct sockaddr_in */
static int name2sockaddrin(char *name, int port, struct sockaddr_in *addrp)
{
    unsigned long inaddr;
    struct hostent *hep;

    bzero(addrp, sizeof(struct sockaddr_in));
    addrp->sin_family = PF_INET;
    if (name != NULL) {
	if ((inaddr = inet_addr(name)) != -1)
	    bcopy(&inaddr, &addrp->sin_addr, sizeof(inaddr));
	else if ((hep = gethostbyname(name)) != NULL)
	    bcopy(hep->h_addr, &addrp->sin_addr, hep->h_length);
	else
	    return (-1);
    }
    else
	addrp->sin_addr.s_addr = htonl(INADDR_ANY);
    addrp->sin_port = htons(port);
    return 0;
}

int
probe_open(char *my_name, char *view_name, int port_no, int ttl)
{
	struct sockaddr_in my_addr;
	int	multicast = 0;
	u_char	c;
	
	/* initialize socket */
	if (name2sockaddrin(my_name, 0, &my_addr) < 0) {
		LogPrintf(LogWARN, "probe: can't get my address!\n");
		return (-1);
	}

	if (name2sockaddrin(view_name, port_no, &view_addr) < 0) {
		LogPrintf(LogWARN, "probe: can't get viewer's address!\n");
		return (1);
	}

#ifdef IN_MULTICAST
	if (IN_MULTICAST(ntohl(view_addr.sin_addr.s_addr)))
		multicast = 1;
#endif
	if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		LogPrintf(LogWARN, "probe: can't open socket");
		return (-1);
	}

#ifdef IP_MULTICAST_TTL
	if (multicast && ttl != 1) {
		c = (u_char)ttl;
		if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_TTL,
			       &c, sizeof(c)) < 0) {
			LogPrintf(LogWARN, "probe: can't set ttl");
			close(sockfd);
			sockfd = -1;
			return (-1);
		}
	}
#endif
	if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
		LogPrintf(LogWARN, "probe: can't bind");
		close(sockfd);
		sockfd = -1;
		return (-1);
	}
	return (0);
}

int
probe_close(void)
{
	return close(sockfd);
}

#define	ROOT_CLASS_HANDLE	0xfffffffe
#define	DEFAULT_CLASS_HANDLE	0xfffffffd

int
probe_send(void)
{
	struct rm_class	*cl;
	struct cbqm_info *info;
	struct class_info *cinfo;
	struct timeval cur_time;
	int	i, len, handle, rval;
	
	gettimeofday(&cur_time, NULL);

	info = (struct cbqm_info *)buffer;
	info->cbqm_magic = CBQM_MAGIC;
	info->cbqm_sec = cur_time.tv_sec;
	info->cbqm_usec = cur_time.tv_usec;

	handle = 0;
	for (i = 0; i < CBQM_MAXCLASS; i++) {
		cl = cbq_classes[i];
		if (cl == NULL)
			break;

		cinfo= &info->cbqm_cinfo[i];

		if (cl == ifd->root_)
			cinfo->handle = ROOT_CLASS_HANDLE;
		else if (cl == ifd->default_)
			cinfo->handle = DEFAULT_CLASS_HANDLE;
		else
			cinfo->handle = handle++;
		cinfo->npackets = cl->stats_.npackets;
		cinfo->nbytes = cl->stats_.nbytes;
		cinfo->drops = cl->stats_.drops;
		cinfo->drop_bytes = cl->stats_.drop_bytes;
	}

	info->cbqm_len = i;

	len = sizeof(struct cbqm_info)
		- (CBQM_MAXCLASS - i) * sizeof(struct class_info);

	if ((rval = sendto(sockfd, buffer, len, 0,
			   (struct sockaddr *)&view_addr,
			   sizeof(view_addr))) < 0)
		LogPrintf(LogWARN, "sendto");
	return (rval);
}

#endif /* CBQ_PROBE */
