/* $Id: altq_cbqcmd.c,v 1.1 1998/06/08 09:56:47 kjc Exp kjc $ */
/*
 * Copyright (C) 1997, 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "opt_altq.h"

#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "command.h"
#include "mbuf.h"
#include "log.h"
#include "loadalias.h"
#include "defs.h"
#include "vars.h"
#include "filter.h"
#include "timer.h"

#include "altq_ppp.h"
#include "altq_classq.h"
#include "altq_rmclass.h"
#include "altq_cbq.h"
#ifdef CBQ_PROBE
#include "../cbqmonitor/cbqmon.h"
#endif

static int cbq_interface_cmd(int argc, char const *const *argv);
static int cbq_class_cmd(int argc, char const *const *argv);
static int cbq_filter_cmd(int argc, char const *const *argv);
static int cbq_probe_cmd(int argc, char const *const *argv);

#define EQUAL(s1, s2)	(strcmp((char *)(s1), (char *)(s2)) == 0)

int
SetCBQ(struct cmdargs const *arg)
{
	if (arg->argc > 0) {
		if (EQUAL(*arg->argv, "interface"))
			return cbq_interface_cmd(arg->argc-1, arg->argv+1);
		else if (EQUAL(*arg->argv, "class"))
			return cbq_class_cmd(arg->argc-1, arg->argv+1);
		else if (EQUAL(*arg->argv, "filter"))
			return cbq_filter_cmd(arg->argc-1, arg->argv+1);
		else if (EQUAL(*arg->argv, "probe"))
			return cbq_probe_cmd(arg->argc-1, arg->argv+1);
		else if (EQUAL(*arg->argv, "enable"))
			return cbq_enable();
		else if (EQUAL(*arg->argv, "disable"))
			return cbq_disable();
		else {
			LogPrintf(LogERROR, "Unknown keyword '%s' for set cbq\n", *arg->argv);
			return (-1);
		}
		
		return 0;
	}
	return -1;
}

int
ReportCBQStatus(struct cmdargs const *arg)
{
	int interval = 0;
	
	if (arg->argc > 0 && EQUAL(*arg->argv, "stop")) {
		cbq_status(-1);
		return (0);
	}

	if (arg->argc > 1) {
		if (EQUAL(*arg->argv, "interval"))
			interval = atoi(*(arg->argv+1));
	}
	cbq_status(interval);
	return (0);
}

static int
cbq_interface_cmd(int argc, char const *const *argv) 
{
	int   	bandwidth = 0;
	int	is_efficient = 1;
	char	*cp;

	altqp->altq_mtu = 1500;	/* set default mtu */

	/*
	 * Look for "bandwidth" & scheduling discipline options.
	 */
	while (argc > 0) {
		if (EQUAL(*argv, "bandwidth")) {
			argc--; argv++;
			cp = NULL;
			bandwidth = strtol(*argv, &cp, 0);
			if (cp != NULL) {
				if (*cp == 'K' || *cp == 'k')
					bandwidth *= 1000;
				if (*cp == 'M' || *cp == 'm')
					bandwidth *= 1000000;
			}
		} else if (EQUAL(*argv, "efficient")) {
			is_efficient = 1;
		} else if (EQUAL(*argv, "-efficient")) {
			is_efficient = 0;
		} else if (EQUAL(*argv, "mtu")) {
			argc--; argv++;
			altqp->altq_mtu = atoi(*argv);
		} else {
			LogPrintf(LogWARN, "Unknown keyword '%s' for cbq interface", *argv);
			return (-1);
		}
		argc--; argv++;
	}

	/* Add this to list of CBQ interfaces */
	if (altq_debug && VarTerm)
		fprintf(VarTerm, "cbq interface: bandwidth=%d, efficient=%d\n",
			bandwidth, is_efficient);

	if (cbq_ifinit(bandwidth, is_efficient)) {
		LogPrintf(LogERROR, "can't init cbq on interface");
		return (-1);
	}

	return (0);
}

static int
cbq_class_cmd(int argc, char const *const *argv) 
{
	char 	class_name[128], parent_name[128];
	u_char  *clname = class_name;
	u_char  *parent = NULL;
	int	borrow = 0;
	u_int   pri = 1;
	int	flags = 0;
	int	pbandwidth = 0;
	int	maxq = 30;
	int	maxburst = 4;
	int	minburst = 1;
	int	av_pkt_size = 0;  /* 0 means use mtu as default */

	/* get class name */
	if (argc > 0) {
		strcpy(class_name, *argv);
		argc--; argv++;
	}

	/* get parent name */
	if (argc > 0) {
		strcpy(parent_name, *argv);
		if (!EQUAL(parent_name, "null") &&
		    !EQUAL(parent_name, "NULL")) {
			parent = parent_name;
		}
		else {
			parent = NULL;
			flags |= CBQCLF_ROOTCLASS;
		}
		argc--; argv++;
	}

	/* look for options. */
	while (argc > 0) {
		if (EQUAL(*argv, "priority")) {
			argc--; argv++;
			pri = atoi(*argv);
		} else if (EQUAL(*argv, "default")) {
			flags |= CBQCLF_DEFCLASS;
		} else if (EQUAL(*argv, "control")) {
			flags |= CBQCLF_CTLCLASS;
		} else if (EQUAL(*argv, "qlength")) {
			argc--; argv++;
			maxq = atoi(*argv);
		} else if (EQUAL(*argv, "borrow")) {
			borrow = 1;
		} else if (EQUAL(*argv, "pbandwidth")) {
			argc--; argv++;
			pbandwidth = atoi(*argv);
		} else if (EQUAL(*argv, "maxburst")) {
			argc--; argv++;
			maxburst = atoi(*argv);
		} else if (EQUAL(*argv, "minburst")) {
			argc--; argv++;
			minburst = atoi(*argv);
		} else if (EQUAL(*argv, "packetsize")) {
			argc--; argv++;
			av_pkt_size = atoi(*argv);
		} else if (EQUAL(*argv, "red")) {
			flags |= CBQCLF_RED;
		} else if (EQUAL(*argv, "ecn")) {
			flags |= (CBQCLF_RED | CBQCLF_ECN);
		} else {
			LogPrintf(LogERROR, "Unknown keyword '%s' for set cbq class\n", *argv);
			return (-1);	
		}
		argc--; argv++;
	}

	if (altq_debug && VarTerm) {
		fprintf(VarTerm, "set class:name=%s,parent=%s,borrow=%d,"
			"pri=%d,pbandwidth=%d\n",
			clname,parent,borrow,pri,pbandwidth);
		fprintf(VarTerm, "    burst=%d,%d,maxq=%d,size=%d,flags=0x%x\n",
			maxburst, minburst, maxq, av_pkt_size, flags);
	}
	if (cbq_class_create(clname, parent, borrow, pri, pbandwidth,
		      maxburst, minburst, maxq, av_pkt_size, flags) == NULL) {
		LogPrintf(LogERROR, "can't add class '%s'\n", clname);
		return (-1);
	}
	return (0);
}


static int
cbq_filter_cmd(int argc, char const *const *argv) 
{
	char 	class_name[128];
	struct rm_class *cl;

	/* get class name */
	if (argc > 0) {
		strcpy(class_name, *argv);
		argc--; argv++;
	}

	if ((cl = cbq_class_find(class_name)) == NULL) {
		LogPrintf(LogERROR, "no cbq class found for a filter\n");
		return (-1);
	}
	
	if (SetCBQfilter(argc, argv, (void *)cl) < 0)
		return (-1);
	
	return (0);
}

#ifdef CBQ_PROBE

static struct pppTimer cbqprobeTimer;

static void
probe_timer(void *data)
{
	int interval;
	
	StopTimer(&cbqprobeTimer);
	probe_send();

	interval = (int)data;
	cbqprobeTimer.load = interval * 1000 / TICKUNIT;
	cbqprobeTimer.func = (void(*)(void *))probe_timer;
	cbqprobeTimer.arg = (void *)interval;
	StartTimer(&cbqprobeTimer);
}

static int
cbq_probe_cmd(int argc, char const *const *argv)
{
	char	*my_name = NULL;
	char	*view_name = NULL;
	int	port_no = CBQM_PORT;
	int	ttl = 1;
	int	interval = 1000;

	while (argc > 0) {
		if (EQUAL(*argv, "stop")) {
			StopTimer(&cbqprobeTimer);
			probe_close();
			return (0);
		}
		else if (EQUAL(*argv, "addr")) {
			argc--; argv++;
			my_name = (char *)*argv;
		}
		else if (EQUAL(*argv, "port")) {
			argc--; argv++;
			port_no = atoi(*argv);
		}
		else if (EQUAL(*argv, "ttl")) {
			argc--; argv++;
			ttl = atoi(*argv);
		}
		else if (EQUAL(*argv, "interval")) {
			argc--; argv++;
			interval = atoi(*argv);
		}
		else if (argc == 1) {
			view_name = (char *)*argv;
		}
		else {
			LogPrintf(LogERROR, "Unknown keyword '%s' for cbq probe\n", *argv);
			return (-1);	
		}
		argc--; argv++;
	}

	if (view_name == NULL) {
		LogPrintf(LogWARN, "cbq probe: missing viewer name\n");
		return (-1);
	}

	if (VarTerm)
		fprintf(VarTerm, "cbq probe: viewer is %s\n", view_name);
	if (probe_open(my_name, view_name, port_no, ttl) < 0) {
		LogPrintf(LogERROR, "probe_open failed!\n");
		return (-1);
	}

	cbqprobeTimer.load = interval * 1000 / TICKUNIT;
	cbqprobeTimer.func = (void(*)(void *))probe_timer;
	cbqprobeTimer.arg = (void *)interval;
	StartTimer(&cbqprobeTimer);

	return (0);
}

#else /* !CBQ_PROBE */

static int
cbq_probe_cmd(int argc, char const *const *argv)
{
	LogPrintf(LogWARN, "cbq probe: not configured\n");
	return (-1);
}
#endif /* !CBQ_PROBE */


