/* $Id: altq_red.c,v 1.1 1998/06/08 09:56:47 kjc Exp kjc $ */
/*
 * Copyright (C) 1997, 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "opt_altq.h"
#ifdef RED	/* red is enabled by RED option in opt_altq.h */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/errno.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "command.h"
#include "mbuf.h"
#include "log.h"
#include "filter.h"
#include "timer.h"
#include "defs.h"
#include "loadalias.h"
#include "vars.h"

#include "altq_ppp.h"
#include "altq_classq.h"
#include "altq_red.h"

/*
 * ALTQ/RED (Random Early Detection) implementation using 32-bit
 * fixed-point calculation.
 *
 * written by kjc using the ns code as a reference.
 * you can learn more about red and ns from Sally's home page at
 * http://www-nrg.ee.lbl.gov/floyd/
 *
 * most of the red parameter values are fixed in this implementation
 * to prevent fixed-point overflow/underflow.
 * if you change the parameters, watch out for overflow/underflow!
 *
 * the parameters used are recommended values by Sally.
 * the corresponding ns config looks:
 *	q_weight=0.00195
 *	minthresh=5 maxthresh=15 queue-size=60
 *	linterm=30
 *	dropmech=random-drop
 *	bytes=false (can't be handled by 32-bit fixed-point)
 *	doubleq=false dqthresh=false 
 *	wait=true
 */
/*
 * alternative red parameters for a slow link.
 *
 * assume the queue length becomes from zero to L and keeps L, it takes
 * N packets for q_avg to reach 63% of L.
 * when q_weight is 0.002, N is about 500 packets.
 * for a slow link like dial-up, 500 packets takes more than 1 minute!
 * when q_weight is 0.008, N is about 127 packets.
 * when q_weight is 0.016, N is about 63 packets.
 * bursts of 50 packets are allowd for 0.002, bursts of 25 packets
 * are allowed for 0.016.
 * see Sally's paper for more details.
 */
/* normal red parameters */
#define W_WEIGHT	512	/* inverse of weight of EWMA (511/512) */
				/* q_weight = 0.00195 */
#define W_SHIFT		9	/* log(W_WEIGHT) */
#define INV_P_MAX	30	/* inverse of max drop probability */

/* red parameters for a slow link */
#define W_WEIGHT_1	128	/* inverse of weight of EWMA (127/128) */
				/* q_weight = 0.0078125 */
#define W_SHIFT_1	7	/* log(W_WEIGHT) */
#define INV_P_MAX_1	20	/* inverse of max drop probability */

/* red parameters for a very slow link (e.g., dialup) */
#define W_WEIGHT_2	64	/* inverse of weight of EWMA (63/64) */
				/* q_weight = 0.015625 */
#define W_SHIFT_2	6	/* log(W_WEIGHT) */
#define INV_P_MAX_2	16	/* inverse of max drop probability */


/* fixed-point uses 12-bit decimal places */
#define FP_SHIFT	12	/* fixed-point shift */

/* fixed red parameters */
#define TH_MIN		5			/* min threshold */
#define TH_MAX		15			/* max threshold */
#define TH_RANGE	(TH_MAX - TH_MIN)	/* threshold range */

#define FP_TH_MIN	(TH_MIN << FP_SHIFT)	/* th_min in fixed-point */

#define RED_LIMIT	60	/* default max queue lenght */
#define RED_STATS		/* collect statistics */

/* internal function prototypes */
static int drop_early __P((red_t *));
static int mark_ecn __P((struct mbuf *, int));
static int pow_w __P((int, int));

/*
 * red support routines
 */

int 
red_init(rp, flags, pkttime)
	red_t 	*rp;
	int	flags;
	int	pkttime;
{
	int npkts_per_sec;
	
	rp->red_avg = 0;
	rp->red_idle = 1;
	rp->red_flags = flags;
	rp->red_pkttime = pkttime;
	if (pkttime == 0)
		rp->red_pkttime = 1;

	rp->red_weight = W_WEIGHT;
	rp->red_wshift = W_SHIFT;
	rp->red_avgshift = W_SHIFT + FP_SHIFT;
	rp->red_inv_pmax = INV_P_MAX;

	/* when the link is very slow, adjust red parameters */
	npkts_per_sec = 1000000 / rp->red_pkttime;
	if (npkts_per_sec < 50) {
		/* up to about 400Kbps */
		rp->red_weight = W_WEIGHT_2;
		rp->red_wshift = W_SHIFT_2;
		rp->red_avgshift = W_SHIFT_2 + FP_SHIFT;
		rp->red_inv_pmax = INV_P_MAX_2;
	}
	else if (npkts_per_sec < 300) {
		/* up to about 2.4Mbps */
		rp->red_weight = W_WEIGHT_1;
		rp->red_wshift = W_SHIFT_1;
		rp->red_avgshift = W_SHIFT_1 + FP_SHIFT;
		rp->red_inv_pmax = INV_P_MAX_1;
	}

	microtime(&rp->red_last);
	return (0);
}

#define DTYPE_NODROP	0	/* no drop */
#define DTYPE_FORCED	1	/* a "forced" drop */
#define DTYPE_EARLY	2	/* an "unforced" (early) drop */

int red_addq(rp, q, m)
	red_t *rp;
	class_queue_t *q;
	struct mbuf *m;
{
	int avg, droptype;
	int n;
    
	avg = rp->red_avg;
    
	/*
	 * if we were idle, we pretend that n packets arrived during
	 * the idle period.
	 */
	if (rp->red_idle) {
		struct timeval now;
		int t;
	
		rp->red_idle = 0;
		microtime(&now);
		t = (now.tv_sec - rp->red_last.tv_sec);
		if (t > 60) {
			/*
			 * being idle for more than 1 minute, set avg to zero.
			 * this prevents t from overflow.
			 */ 
			avg = 0;
		}
		else {
			t = t * 1000000 + (now.tv_usec - rp->red_last.tv_usec);
			n = t / rp->red_pkttime;

			/* the following line does (avg = (1 - Wq)^n * avg) */
			if (n > 0)
				avg = (avg >> FP_SHIFT) *
					pow_w(rp->red_weight, n);
		}
	}

	/* run estimator. (note: avg is scaled by 512 in fixed-point) */
	avg += (qlen(q) << FP_SHIFT) - (avg >> rp->red_wshift);
	rp->red_avg = avg;		/* save the new value */

	/*
	 * red_count keeps a tally of arriving traffic that has not
	 * been dropped.
	 */
	rp->red_count++;
    
	/* see if we drop early */
	droptype = DTYPE_NODROP;
	if (avg >= (TH_MIN << rp->red_avgshift) && qlen(q) > 1) {
		if (avg >= (TH_MAX << rp->red_avgshift)) {
			/* avg >= th_max: forced drop */
			droptype = DTYPE_FORCED;
		}
		else if (rp->red_old == 0) {
			/* first exceeds th_min */
			rp->red_count = 1;
			rp->red_old = 1;
		}
		else if (drop_early(rp)) {
			/* mark or drop by red */
			if ((rp->red_flags & REDF_ECN) &&
			    mark_ecn(m, rp->red_flags)) {
				/* successfully marked.  do not drop. */
				rp->red_count = 0;
#ifdef RED_STATS
				rp->red_stats.marked_packets++;
#endif
			}
			else { 
				/* unforced drop by red */
				droptype = DTYPE_EARLY;
			}
		}
	}
	else {
		/* avg < th_min */
		rp->red_old = 0;
	}

	/*
	 * if the queue length hits the hard limit, it's a forced drop.
	 */
	if (droptype == DTYPE_NODROP && qlen(q) >= qlimit(q))
		droptype = DTYPE_FORCED;

	/* if successful or forced drop, enqueue this packet. */
	if (droptype != DTYPE_EARLY)
		_addq(q, m);

	if (droptype != DTYPE_NODROP) {
		if (droptype == DTYPE_EARLY) {
			/* drop the incoming packet */
#ifdef RED_STATS
			rp->red_stats.drop_unforced++;
#endif
		}
		else {
			/* forced drop, select a victim packet in the queue. */
			m = _getq_random(q);
#ifdef RED_STATS
			rp->red_stats.drop_forced++;
#endif
		}
#ifdef RED_STATS
		rp->red_stats.drop_packets++;
		rp->red_stats.drop_bytes += m_pktlen(m);
#endif
		rp->red_count = 0;
		m_freem(m);
		return (-1);
	}
	/* successfully queued */
	return (0);
}

/*
 * early-drop probability is calculated as follows:
 * prob = p_max * (avg - th_min) / (th_max - th_min)
 * prob_a = prob / (2 - count*prob)
 *	  = (avg-th_min) / (2*(th_max-th_min)*inv_p_max - count*(avg-th_min))
 *
 */
static int drop_early(rp)
	red_t *rp;
{
	int avg;	/* avg in fixed-point */
	int d;		/* denominator of drop-probability */

	avg = rp->red_avg >> rp->red_wshift;	/* unscale avg */

	d = ((2 * TH_RANGE * rp->red_inv_pmax) << FP_SHIFT)
		- rp->red_count * (avg - FP_TH_MIN);

	if (d <= 0)
		/* count exceeds the hard limit: drop or mark */
		return (1);

	/*
	 * now the range of d is [1..600] in fixed-point.
	 *
	 * drop probability = (avg - FP_TH_MIN) / d
	 */

	if ((random() % d) < (avg - FP_TH_MIN)) {
		/* drop or mark */
		return (1);
	}
	/* no drop/mark */
	return (0);
}

/*
 * try to mark CE bit to the packet.
 *    returns 1 if successfully marked, 0 otherwise.
 */
#ifndef IPTOS_CE
#define IPTOS_CE		0x01	/* congestion experienced */
#define IPTOS_ECT		0x02	/* ECN-capable transport */
#endif

static int mark_ecn(m, flags)
	struct mbuf *m;
	int flags;
{
	if (flags & REDF_ECN4) {
		struct ip *ip = mtod(m, struct ip *);
	    
		if (ip->ip_tos & IPTOS_ECT) {
			/* ECN-capable, mark ECN bit. */
			if ((ip->ip_tos & IPTOS_CE) == 0) {
				long sum;
		    
				ip->ip_tos |= IPTOS_CE;
				/*
				 * update checksum (from RFC1624)
				 *	   HC' = ~(~HC + ~m + m')
				 */
				sum = ~ntohs(ip->ip_sum) & 0xffff;
				sum += 0xffff + IPTOS_CE;
				sum = (sum >> 16) + (sum & 0xffff);
				sum += (sum >> 16);  /* add carry */
		
				ip->ip_sum = htons(~sum & 0xffff);
			}
			return (1);
		}
	}

	/* not marked */
	return (0);
}

struct mbuf *red_getq(rp, q)
	red_t *rp;
	class_queue_t *q;
{
	struct mbuf *m;
	
	if ((m = _getq(q)) == NULL) {
		if (rp->red_idle == 0) {
			rp->red_idle = 1;
			microtime(&rp->red_last);
		}
		return NULL;
	}

	rp->red_idle = 0;
#ifdef RED_STATS
	rp->red_stats.xmit_packets++;
	rp->red_stats.xmit_bytes += m_pktlen(m);
#endif
	return (m);
}

/*
 * helper routine to calibrate avg during idle.
 * it currently supports only three weight values.
 * pow_w(w, n) returns (1 - Wq)^n in fixed-point
 * here Wq = 1/w and the code assumes Wq is close to zero.
 *
 * wtab[n] holds ((1 - Wq)^(2^n)) in fixed-point.
 */
static int32_t wtable[3][32];
static int w_param_max[3] = {0, 0, 0};/* max effective parameter for pow_w() */

static int32_t pow_w(w, n)
	int w, n;
{
	int index, i, bit;
	int32_t *wtab, val;

	if (w == W_WEIGHT)
		index = 0;
	else if (w == W_WEIGHT_1)
		index = 1;
	else if (w == W_WEIGHT_2)
		index = 2;
	else {
		LogPrintf(LogERROR, "red: bad weight %d", w);
		index = 2;
	}

	wtab = &wtable[index][0];
	
	if (w_param_max[index] == 0) {
		/* first time called. initialize wtab. */
		wtab[0] = ((w - 1) << FP_SHIFT) / w;
		for (i = 1; i < 32; i++) {
			wtab[i] = (wtab[i-1] * wtab[i-1]) >> FP_SHIFT;
			if (wtab[i] == 0 && w_param_max[index] == 0)
				w_param_max[index] = 1 << i;
		}
	}
    
	if (n >= w_param_max[index])
		return (0);

	val = 1 << FP_SHIFT;
	if (n <= 0)
		return (val);

	bit = 1;
	i = 0;
	while (n) {
		if (n & bit) {
			val = (val * wtab[i]) >> FP_SHIFT;
			n &= ~bit;
		}
		i++;
		bit <<=  1;
	}
	return (val);
}

#endif /* RED */
